"""Write Value Change Dump files.

This module provides :class:`VCDWriter` for writing VCD files.

"""
from datetime import datetime
from itertools import zip_longest
from numbers import Number
from types import TracebackType
from typing import (
    IO,
    Dict,
    Generator,
    Generic,
    List,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
)

from vcd.common import ScopeType, Timescale, TimescaleMagnitude, TimescaleUnit, VarType


class VCDPhaseError(Exception):
    """Indicating a :class:`VCDWriter` method was called in the wrong phase.

    For example, calling :meth:`register_var()` after :meth:`close()` will raise this
    exception.

    """


ScopeTuple = Tuple[str, ...]
ScopeInput = Union[str, Sequence[str]]
TimeValue = Union[int, float]
TimescaleLike = Union[Timescale, Tuple[int, str], str]
CompoundSize = Sequence[int]
VariableSize = Union[int, CompoundSize]

EventValue = Union[bool, int]
RealValue = Union[float, int]
ScalarValue = Union[int, bool, str, None]
StringValue = Union[str, None]
CompoundValue = Sequence[ScalarValue]
VarValue = Union[EventValue, RealValue, ScalarValue, StringValue, CompoundValue]


class VCDWriter:
    """Value Change Dump writer.

    A VCD file captures time-ordered changes to the value of variables.

    :param file file: A file-like object to write the VCD data.
    :param timescale:
        Scale of the VCD timestamps. The timescale may either be a string or a tuple
        containing an (int, str) pair.
    :type timescale: str, tuple
    :param str date: Optional `$date` string used in the VCD header.
    :param str comment: Optional `$comment` string used in the VCD header.
    :param str version: Optional `$version` string used in the VCD header.
    :param str default_scope_type: Scope type for scopes where :meth:`set_scope_type()`
            is not called explicitly.
    :param str scope_sep: Separator for scopes specified as strings.
    :param int init_timestamp: The initial timestamp. default=0
    :raises ValueError: for invalid timescale values

    """

    def __init__(
        self,
        file: IO[str],
        timescale: TimescaleLike = '1 us',
        date: Optional[str] = None,
        comment: str = '',
        version: str = '',
        default_scope_type: Union[ScopeType, str] = ScopeType.module,
        scope_sep: str = '.',
        check_values: bool = True,
        init_timestamp: TimeValue = 0,
    ) -> None:
        self._ofile = file
        self._header_keywords = {
            '$timescale': self._check_timescale(timescale),
            '$date': str(datetime.now()) if date is None else date,
            '$comment': comment,
            '$version': version,
        }
        self._default_scope_type = ScopeType(default_scope_type)
        self._scope_sep = scope_sep
        self._check_values = check_values
        self._registering = True
        self._closed = False
        self._dumping = True
        self._next_var_id: int = 1
        self._scope_var_strs: Dict[ScopeTuple, List[str]] = {}
        self._scope_var_names: Dict[ScopeTuple, Set[str]] = {}
        self._scope_types: Dict[ScopeTuple, ScopeType] = {}
        self._vars: List[Variable] = []
        self._timestamp = int(init_timestamp)
        self._last_dumped_ts: Optional[int] = None

    def set_scope_type(
        self, scope: ScopeInput, scope_type: Union[ScopeType, str]
    ) -> None:
        """Set the scope_type for a given scope.

        The scope's type may be set to one of the valid :class:`ScopeType` values. VCD
        viewer applications may display different scope types differently.

        :param scope: The scope to set the type of.
        :type scope: str or sequence of str
        :param str scope_type: A valid scope type string.
        :raises ValueError: for invalid `scope_type`

        """
        scope_type = ScopeType(scope_type)
        scope_tuple = self._get_scope_tuple(scope)
        self._scope_types[scope_tuple] = scope_type

    def register_var(
        self,
        scope: ScopeInput,
        name: str,
        var_type: Union[VarType, str],
        size: Optional[VariableSize] = None,
        init: VarValue = None,
    ) -> 'Variable':
        """Register a new VCD variable.

        All VCD variables must be registered prior to any value changes.

        :param scope: The hierarchical scope that the variable belongs within.
        :type scope: str or sequence of str
        :param str name: Name of the variable.
        :param VarType var_type: Type of the variable.
        :param size:
            Size, in bits, of the variable. The *size* may be expressed as an int or,
            for vector variable types, a tuple of int. When the size is expressed as a
            tuple, the *value* passed to :meth:`change()` must also be a tuple of same
            arity as the *size* tuple. Some variable types ('integer', 'real',
            'realtime', and 'event') have a default size and thus *size* may be ``None``
            for those variable types.
        :type size: int or tuple(int) or None
        :param init: Optional initial value; defaults to 'x'.
        :raises VCDPhaseError: if any values have been changed
        :raises ValueError: for invalid var_type value
        :raises TypeError: for invalid parameter types
        :raises KeyError: for duplicate var name
        :returns: :class:`Variable` instance appropriate for use with :meth:`change()`.

        """
        if self._closed:
            raise VCDPhaseError('Cannot register after close().')
        elif not self._registering:
            raise VCDPhaseError('Cannot register after time 0.')
        var_type = VarType(var_type)

        scope_tuple = self._get_scope_tuple(scope)

        scope_names = self._scope_var_names.setdefault(scope_tuple, set())
        if name in scope_names:
            raise KeyError(
                f'Duplicate var {name} in scope {self._scope_sep.join(scope_tuple)}'
            )

        if size is None:
            if var_type in [VarType.integer, VarType.real, VarType.realtime]:
                size = 64
            elif var_type in [VarType.event, VarType.string]:
                size = 1
            else:
                raise ValueError(f'Must supply size for {var_type} var_type')

        if isinstance(size, Sequence):
            size = tuple(size)
            var_size = sum(size)
        else:
            var_size = size

        ident = _encode_identifier(self._next_var_id)

        var_str = f'$var {var_type} {var_size} {ident} {name} $end'

        var: Variable
        if var_type == VarType.string:
            if init is None:
                init = ''
            elif not isinstance(init, str):
                raise ValueError('string init value must be str')
            var = StringVariable(ident, var_type, size, init)
        elif var_type == VarType.event:
            if init is None:
                init = True
            elif not isinstance(init, (bool, int)):
                raise ValueError('event init value must be int, bool, or None')
            var = EventVariable(ident, var_type, size, init)
        elif var_type == VarType.real:
            if init is None:
                init = 0.0
            elif not isinstance(init, (float, int)):
                raise ValueError('real init value must be float, int, or None')
            var = RealVariable(ident, var_type, size, init)
        elif size == 1:
            if init is None:
                init = 'x'
            elif not isinstance(init, (int, bool, str)):
                raise ValueError('scalar init value must be int, bool, str, or None')
            var = ScalarVariable(ident, var_type, size, init)
        elif isinstance(size, tuple):
            if init is None:
                init = tuple('x' * len(size))
            elif not isinstance(init, Sequence):
                raise ValueError('compount init value must be a sequence')
            elif len(init) != len(size):
                raise ValueError('compound init value must be same length as size')
            elif not all(isinstance(v, (int, bool, str)) for v in init):
                raise ValueError('compound init values must be int, bool, or str')
            var = CompoundVectorVariable(ident, var_type, size, init)
        else:
            if init is None:
                init = 'x'
            elif not isinstance(init, (int, bool, str)):
                raise ValueError('vector init value must be int, bool, str, or None')
            var = VectorVariable(ident, var_type, size, init)

        var.format_value(init, check=True)

        # Only alter state after format_value() succeeds
        self._vars.append(var)
        self._next_var_id += 1
        self._scope_var_strs.setdefault(scope_tuple, []).append(var_str)
        scope_names.add(name)

        return var

    def register_alias(self, scope: ScopeInput, name: str, var: 'Variable') -> None:
        """Register a variable alias.

        The same VCD identifier may be associated with multiple reference names ("$var"
        declarations). This method associates an existing :class:`Variable` instance
        with a different variable scope and/or name. The alias shares the same
        identifier, type, size, and value as the reference variable. Because the
        identifier is shared, calling :meth:`change()` with ``var`` changes the value of
        of all associated reference names.

        :param scope: The hierarchical scope that the variable belongs within.
        :type scope: str or sequence of str
        :param str name: Name of the variable.
        :param Variable var: Existing variable to alias.

        """
        if self._closed:
            raise VCDPhaseError('Cannot register after close().')
        elif not self._registering:
            raise VCDPhaseError('Cannot register after time 0.')

        scope_tuple = self._get_scope_tuple(scope)
        scope_names = self._scope_var_names.setdefault(scope_tuple, set())
        if name in scope_names:
            raise KeyError(
                f'Duplicate var {name} in scope {self._scope_sep.join(scope_tuple)}'
            )

        var_str = f'$var {var.type} {var.size} {var.ident} {name} $end'
        self._scope_var_strs.setdefault(scope_tuple, []).append(var_str)
        scope_names.add(name)

    def dump_off(self, timestamp: TimeValue) -> None:
        """Suspend dumping to VCD file."""
        if self._registering:
            self._finalize_registration()
        self._set_timestamp(timestamp)
        if not self._dumping:
            return
        self._dump_timestamp()
        self._ofile.write('$dumpoff\n')
        for var in self._vars:
            val_str = var.dump_off()
            if val_str:
                self._ofile.write(val_str + '\n')
        self._ofile.write('$end\n')
        self._dumping = False

    def dump_on(self, timestamp: TimeValue) -> None:
        """Resume dumping to VCD file."""
        if self._registering:
            self._finalize_registration()
        self._set_timestamp(timestamp)
        if self._dumping:
            return
        self._dumping = True
        self._dump_timestamp()
        self._dump_values('$dumpon')

    def _dump_values(self, keyword: str) -> None:
        self._ofile.write(keyword + '\n')
        for var in self._vars:
            val_str = var.dump(self._check_values)
            if val_str:
                self._ofile.write(val_str + '\n')
        self._ofile.write('$end\n')

    def _set_timestamp(self, timestamp: TimeValue) -> None:
        if timestamp < self._timestamp:
            raise VCDPhaseError(f'Out of order timestamp: {timestamp}')
        elif timestamp > self._timestamp:
            self._timestamp = int(timestamp)

    def _dump_timestamp(self) -> None:
        if (self._timestamp != self._last_dumped_ts and self._dumping) or (
            self._last_dumped_ts is None
        ):
            self._last_dumped_ts = self._timestamp
            self._ofile.write(f'#{self._timestamp}\n')

    def change(self, var: 'Variable', timestamp: TimeValue, value: VarValue) -> None:
        """Change variable's value in VCD stream.

        This is the fundamental behavior of a :class:`VCDWriter` instance. Each time a
        variable's value changes, this method should be called.

        The *timestamp* must be in-order relative to timestamps from previous calls to
        :meth:`change()`. It is okay to call :meth:`change()` multiple times with the
        same *timestamp*, but never with a past *timestamp*.

        .. Note::

            :meth:`change()` may be called multiple times before the timestamp
            progresses past 0. The last value change for each variable will go into the
            $dumpvars section.

        :param Variable var: :class:`Variable` instance (i.e. from
                             :meth:`register_var()`).
        :param int timestamp: Current simulation time.
        :param value:
            New value for *var*. For :class:`VectorVariable`, if the variable's *size*
            is a tuple, then *value* must be a tuple of the same arity.

        :raises ValueError: if the value is not valid for *var*.
        :raises VCDPhaseError: if the timestamp is out of order or the
                               :class:`VCDWriter` instance is closed.

        """
        if self._closed:
            raise VCDPhaseError('Cannot change value after close()')

        # Format value early to catch any errors before writing output.
        if value != var.value or var.type == VarType.event:
            val_str = var.format_value(value, self._check_values)
        else:
            val_str = ''

        # Unroll for performance: self._set_timestamp(timestamp)
        if timestamp < self._timestamp:
            raise VCDPhaseError(f'Out of order timestamp: {timestamp}')
        elif timestamp > self._timestamp:
            if self._registering:
                self._finalize_registration()
            self._timestamp = int(timestamp)

        if not val_str:
            return

        var.value = value
        if self._dumping and not self._registering:
            # Unroll for performance: self._dump_timestamp()
            if self._timestamp != self._last_dumped_ts:
                self._last_dumped_ts = self._timestamp
                self._ofile.write(f'#{self._timestamp}\n{val_str}\n')
            else:
                self._ofile.write(f'{val_str}\n')

    def _get_scope_tuple(self, scope: ScopeInput) -> ScopeTuple:
        if isinstance(scope, str):
            return tuple(scope.split(self._scope_sep))
        if isinstance(scope, Sequence):
            return tuple(scope)
        else:
            raise TypeError(f'Invalid scope {scope}')

    @classmethod
    def _check_timescale(cls, timescale: TimescaleLike) -> str:
        if isinstance(timescale, Timescale):
            return str(timescale)
        elif isinstance(timescale, (list, tuple)):
            if len(timescale) != 2:
                raise ValueError(f'Invalid timescale {timescale}')
            mag, unit = timescale
            return str(Timescale(TimescaleMagnitude(mag), TimescaleUnit(unit)))
        elif isinstance(timescale, str):
            return str(Timescale.from_str(timescale))
        else:
            raise TypeError(f'Invalid timescale type {type(timescale).__name__}')

    def __enter__(self) -> 'VCDWriter':
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[Exception]],
        exc_value: Optional[Exception],
        traceback: Optional[TracebackType],
    ) -> None:
        self.close()

    def close(self, timestamp: Optional[TimeValue] = None) -> None:
        """Close VCD writer.

        Any buffered VCD data is flushed to the output file. After :meth:`close()`, no
        variable registration or value changes will be accepted.

        :param int timestamp: optional final timestamp to insert into VCD stream.

        .. Note::

            The output file is not automatically closed. It is up to the user to ensure
            the output file is closed after the :class:`VCDWriter` instance is closed.

        """
        if not self._closed:
            self.flush(timestamp)
            self._closed = True

    def flush(self, timestamp: Optional[TimeValue] = None) -> None:
        """Flush any buffered VCD data to output file.

        If the VCD header has not already been written, calling `flush()` will force the
        header to be written thus disallowing any further variable registration.

        :param int timestamp: optional timestamp to insert into VCD stream.

        """
        if self._closed:
            raise VCDPhaseError('Cannot flush() after close()')
        if self._registering:
            self._finalize_registration()
        if timestamp is not None:
            self._set_timestamp(timestamp)
            self._dump_timestamp()
        self._ofile.flush()

    def _gen_header(self) -> Generator[str, None, None]:
        for kwname, kwvalue in sorted(self._header_keywords.items()):
            if not kwvalue:
                continue
            lines = kwvalue.split('\n')
            if len(lines) == 1:
                yield f'{kwname} {lines[0]} $end'
            else:
                yield kwname
                for line in lines:
                    yield '\t' + line
                yield '$end'

        prev_scope: ScopeTuple = ()
        for scope in sorted(self._scope_var_strs):
            var_strs = self._scope_var_strs.pop(scope)

            for i, (prev, this) in enumerate(zip_longest(prev_scope, scope)):
                if prev != this:
                    for _ in prev_scope[i:]:
                        yield '$upscope $end'

                    for j, name in enumerate(scope[i:]):
                        scope_type = self._scope_types.get(
                            scope[: i + j + 1], self._default_scope_type
                        )
                        yield f'$scope {scope_type.value} {name} $end'
                    break
            else:
                assert scope != prev_scope  # pragma no cover

            for var_str in var_strs:
                yield var_str

            prev_scope = scope

        for _ in prev_scope:
            yield '$upscope $end'

        yield '$enddefinitions $end'

    def _finalize_registration(self) -> None:
        assert self._registering
        self._ofile.write('\n'.join(self._gen_header()) + '\n')
        if self._vars:
            self._dump_timestamp()
            self._dump_values('$dumpvars')
        self._registering = False

        # This state is not needed after registration phase.
        self._header_keywords.clear()
        self._scope_types.clear()
        self._scope_var_names.clear()


ValueType = TypeVar('ValueType')


class Variable(Generic[ValueType]):
    """VCD variable details needed to call :meth:`VCDWriter.change()`."""

    __slots__ = ('ident', 'type', 'size', 'value')

    def __init__(self, ident: str, type: VarType, size: VariableSize, init: ValueType):
        #: Identifier used in VCD output stream.
        self.ident = ident
        #: VCD variable type; one of :const:`VCDWriter.VAR_TYPES`.
        self.type = type
        #: Size, in bits, of variable.
        self.size = size
        #: Last value of variable.
        self.value = init

    def format_value(self, value: ValueType, check: bool = True) -> str:
        """Format value change for use in VCD stream."""
        raise NotImplementedError

    def dump(self, check: bool = True) -> Optional[str]:
        return self.format_value(self.value, check)

    def dump_off(self) -> Optional[str]:
        return None


class ScalarVariable(Variable[ScalarValue]):
    """One-bit VCD scalar.

    This is a 4-state variable and thus may have values of 0, 1, 'z', or 'x'.

    """

    __slots__ = ()

    def format_value(self, value: ScalarValue, check: bool = True) -> str:
        """Format scalar value change for VCD stream.

        :param value: 1-bit (4-state) scalar value.
        :type value: str, bool, int, or None
        :raises ValueError: for invalid *value*.
        :returns: string representing value change for use in a VCD stream.

        """
        if isinstance(value, str):
            if check and (len(value) != 1 or value not in '01xzXZ'):
                raise ValueError(f'Invalid scalar value ({value})')
            return value + self.ident
        elif value is None:
            return 'z' + self.ident
        elif value:
            return '1' + self.ident
        else:
            return '0' + self.ident

    def dump_off(self) -> str:
        return 'x' + self.ident


class EventVariable(Variable[EventValue]):
    """VCD event variable.

    An event is transient--it only exists at the time it is changed.

    """

    def format_value(self, value: EventValue, check: bool = True) -> str:
        if value:
            return '1' + self.ident
        else:
            raise ValueError('Invalid event value')

    def dump(self, check: bool = True) -> Optional[str]:
        return None


class StringVariable(Variable[StringValue]):
    """String variable as known by GTKWave.

    Any "string" (character-chain) can be displayed as a change. This type is only
    supported by GTKWave.

    """

    __slots__ = ()

    def format_value(self, value: StringValue, check: bool = True) -> str:
        """Format scalar value change for VCD stream.

        :param value: a string, str()
        :type value: str
        :raises ValueError: for invalid *value*.
        :returns: string representing value change for use in a VCD stream.

        """
        if value is None:
            value_str = ''
        elif check and not isinstance(value, str):
            raise ValueError(f'Invalid string value ({value!r})')
        else:
            value_str = value.translate(
                {
                    9: "\\t",
                    10: "\\n",
                    13: "\\r",
                    32: "\\x20",
                    92: "\\\\",
                }
            )
        return f's{value_str} {self.ident}'


class RealVariable(Variable[RealValue]):
    """Real (IEEE-754 double-precision floating point) variable.

    Values must be numeric and cannot be 'x' or 'z' states.

    """

    __slots__ = ()

    def format_value(self, value: RealValue, check: bool = True) -> str:
        """Format real value change for VCD stream.

        :param value: Numeric changed value.
        :param type: float or int
        :raises ValueError: for invalid real *value*.
        :returns: string representing value change for use in a VCD stream.

        """
        if not check or isinstance(value, Number):
            return f'r{value:.16g} {self.ident}'
        else:
            raise ValueError(f'Invalid real value ({value})')


class VectorVariable(Variable[ScalarValue]):
    """Bit vector variable type.

    This is for the various non-scalar and non-real variable types including integer,
    register, wire, etc.

    """

    __slots__ = ()

    size: int

    def format_value(self, value: ScalarValue, check: bool = True) -> str:
        """Format value change for VCD stream.

        :param value: New value for the variable.
        :types value: int, str, or None
        :raises ValueError: for *some* invalid values.

        A *value* of `None` is the same as `'z'`.

        .. Warning::

            If *value* is of type :py:class:`str`, all characters must be one of
            `'01xzXZ'`. For the sake of performance, checking **is not** done to ensure
            value strings only contain conforming characters. Thus it is possible to
            produce invalid VCD streams with invalid string values.

        """
        value_str = _format_scalar_value(value, self.size, check)
        return f'b{value_str} {self.ident}'

    def dump_off(self) -> str:
        return self.format_value('x', check=False)


class CompoundVectorVariable(Variable[CompoundValue]):
    """Bit vector variable type with a compound size.

    This is for the various non-scalar and non-real variable types including integer,
    register, wire, etc.

    """

    __slots__ = ()

    size: CompoundSize

    def format_value(self, value: CompoundValue, check: bool = True) -> str:
        """Format value change for VCD stream.

        :param value: Sequence of scalar components of the variable's value. The
                      sequence must be the same length as the variable's size tuple.
        :returns: string representing value change for use in a VCD stream.

        """
        if len(value) != len(self.size):
            raise ValueError(
                f'Compound value ({value}) must be length {len(self.size)}'
            )
        # The string is built-up right-to-left in order to minimize/avoid left-extension
        # in the final value string.
        vstr_list: List[str] = []
        vstr_len = 0
        size_sum = 0
        for v, size in zip(reversed(value), reversed(self.size)):
            vstr = _format_scalar_value(v, size, check)
            if not vstr_list:
                vstr_list.insert(0, vstr)
                vstr_len += len(vstr)
            else:
                leftc = vstr_list[0][0]
                rightc = vstr[0]
                if len(vstr) > 1 or (
                    (rightc != leftc or leftc == '1')
                    and (rightc != '0' or leftc != '1')
                ):
                    extendc = '0' if leftc == '1' else leftc
                    extend_size = size_sum - vstr_len
                    vstr_list.insert(0, extendc * extend_size)
                    vstr_list.insert(0, vstr)
                    vstr_len += extend_size + len(vstr)
            size_sum += size
        value_str = ''.join(vstr_list)
        return f'b{value_str} {self.ident}'

    def dump_off(self) -> str:
        return self.format_value(tuple('x' * len(self.size)), check=False)


def _format_scalar_value(value: ScalarValue, size: int, check: bool) -> str:
    if isinstance(value, int):
        max_val = 1 << size
        if check and (-value > (max_val >> 1) or value >= max_val):
            raise ValueError(f'Value ({value}) not representable in {size} bits')
        if value < 0:
            value += max_val
        return format(value, 'b')
    elif value is None:
        return 'z'
    else:
        if check and (
            not isinstance(value, str)
            or len(value) > size
            or any(c not in '01xzXZ-' for c in value)
        ):
            raise ValueError(f'Invalid vector value ({value})')
        return value


def _encode_identifier(v: int) -> str:
    """Encode identifer value into base-94 string."""
    assert v > 0, 'identifier codes must be > 0'
    encoded = ''
    while v != 0:
        v -= 1
        encoded += chr((v % 94) + 33)
        v //= 94
    return encoded



'''Read Value Change Dump (VCD) files.

The primary interface is the :func:`tokenize()` generator function,
parses a binary VCD stream, yielding tokens as they are encountered.

.. code::

   >>> import io
   >>> from vcd.reader import TokenKind, tokenize
   >>> vcd = b"$date today $end $timescale 1 ns $end"
   >>> tokens = tokenize(io.BytesIO(vcd))
   >>> token = next(tokens)
   >>> assert token.kind is TokenKind.DATE
   >>> assert token.date == 'today'
   >>> token = next(tokens)
   >>> assert token.kind is TokenKind.TIMESCALE
   >>> assert token.timescale.magnitude.value == 1
   >>> assert token.timescale.unit.value == 'ns'

'''

import io
from dataclasses import dataclass
from enum import Enum
from typing import Iterator, NamedTuple, Optional, Tuple, Union

from vcd.common import ScopeType, Timescale, TimescaleMagnitude, TimescaleUnit, VarType


class TokenKind(Enum):
    """Kinds of VCD tokens."""

    COMMENT = 1
    DATE = 2
    ENDDEFINITIONS = 3
    SCOPE = 4
    TIMESCALE = 5
    UPSCOPE = 6
    VAR = 7
    VERSION = 8
    DUMPALL = 9
    DUMPOFF = 10
    DUMPON = 11
    DUMPVARS = 12
    END = 13
    CHANGE_TIME = 14
    CHANGE_SCALAR = 15
    CHANGE_VECTOR = 16
    CHANGE_REAL = 17
    CHANGE_STRING = 18


class VarDecl(NamedTuple):
    """VCD variable declaration.

    Examples::

       $var wire 4 !@# foobar [ 3 : 1 ] $end
       $var real 1 aaa foobar $end
       $var integer 32 > foobar[8] $end

    """

    type_: VarType  #: Type of variable
    size: int  #: Size, in bits, of variable
    id_code: str
    """Identifer code of variable.

    This code is used in subsequent value change descriptors
    to map-back to this variable declaration."""

    reference: str
    """Reference name of variable.

    This human-readable name typically corresponds to the name of a
    variable in the model that output the VCD."""

    bit_index: Union[None, int, Tuple[int, int]]
    """Optional range of bits to select from the variable.

    May select a single bit index, e.g. ``ref [ 3 ]``. Or a range of
    bits, e.g. from ``ref [ 7 : 3 ]`` (MSB index then LSB index)."""

    @property
    def ref_str(self) -> str:
        if self.bit_index is None:
            return self.reference
        elif isinstance(self.bit_index, int):
            return f'{self.reference}[{self.bit_index}]'
        else:
            return f'{self.reference}[{self.bit_index[0]}:{self.bit_index[1]}]'


class ScopeDecl(NamedTuple):
    """VCD scope declaration.

    Examples::

       $scope module Foo $end
       $scope
          fork alpha_beta
       $end

    """

    type_: ScopeType  #: Type of scope
    ident: str  #: Scope name


class VectorChange(NamedTuple):
    """Vector value change descriptor.

    A vector value consists of multiple 4-state values, where the four
    states are 0, 1, X, and Z. When a vector value consists entirely
    of 0 and 1 states, :attr:`value` will be an int. Otherwise
    :attr:`value` will be a str.

    """

    id_code: str  #: Identifier code of associated variable.
    value: Union[int, str]  #: New value of associated vector variable.


class RealChange(NamedTuple):
    """Real value (floating point) change descriptor."""

    id_code: str  #: Identifier code of associated variable.
    value: float  #: New value of associated real variable.


class ScalarChange(NamedTuple):
    """Scalar value change descriptor.

    A scalar is a single 4-state value. The value is one of '0', '1',
    'X', or 'Z'.

    """

    id_code: str  #: Identifier code of associated variable.
    value: str  #: New value of associated scalar variable.


class StringChange(NamedTuple):
    """String value change descriptor.

    Strings are VCD extension supported by GTKWave.

    """

    id_code: str  #: Identifier code of associated variable.
    value: str  #: New value of associated string variable.


class Location(NamedTuple):
    """Describe location within VCD stream/file."""

    line: int  #: Line number
    column: int  #: Column number


class Span(NamedTuple):
    """Describe location span within VCD stream/file."""

    start: Location  #: Start of span
    end: Location  #: End of span


class Token(NamedTuple):
    """VCD token yielded from :func:`tokenize()`.

    These are relatively high-level tokens insofar as each token fully
    captures an entire VCD declaration, command, or change descriptor.

    The :attr:`kind` attribute determines the :attr:`data` type. Various
    kind-specific properties provide runtime type-checked access to the
    kind-specific data.

    .. Note::

       The :attr:`data` attribute may be accessed directly to avoid
       runtime type checks and thus achieve better runtime performance
       versus accessing kind-specific properties such as
       :attr:`scalar_change`.

    """

    kind: TokenKind
    "The kind of token."

    span: Span
    "The start and end location of the token within the file/stream."

    data: Union[
        None,  # $enddefinitions $upscope $dump* $end
        int,  # time change
        str,  # $comment, $date, $version
        ScopeDecl,  # $scope
        Timescale,  # $timescale
        VarDecl,  # $var
        ScalarChange,
        VectorChange,
        RealChange,
        StringChange,
    ]
    "Data associated with the token. The data type depends on :attr:`kind`."

    @property
    def comment(self) -> str:
        """Unstructured text from a ``$comment`` declaration."""
        assert self.kind is TokenKind.COMMENT
        assert isinstance(self.data, str)
        return self.data

    @property
    def date(self) -> str:
        """Unstructured text from a ``$date`` declaration."""
        assert self.kind is TokenKind.DATE
        assert isinstance(self.data, str)
        return self.data

    @property
    def scope(self) -> ScopeDecl:
        """Scope type and identifier from ``$scope`` declaration."""
        assert self.kind is TokenKind.SCOPE
        assert isinstance(self.data, ScopeDecl)
        return self.data

    @property
    def timescale(self) -> Timescale:
        """Magnitude and unit from ``$timescale`` declaration."""
        assert self.kind is TokenKind.TIMESCALE
        assert isinstance(self.data, Timescale)
        return self.data

    @property
    def var(self) -> VarDecl:
        """Details from a ``$var`` declaration."""
        assert self.kind is TokenKind.VAR
        assert isinstance(self.data, VarDecl)
        return self.data

    @property
    def version(self) -> str:
        """Unstructured text from a ``$version`` declaration."""
        assert self.kind is TokenKind.VERSION
        assert isinstance(self.data, str)
        return self.data

    @property
    def time_change(self) -> int:
        """Simulation time change."""
        assert self.kind is TokenKind.CHANGE_TIME
        assert isinstance(self.data, int)
        return self.data

    @property
    def scalar_change(self) -> ScalarChange:
        """Scalar value change descriptor."""
        assert self.kind is TokenKind.CHANGE_SCALAR
        assert isinstance(self.data, ScalarChange)
        return self.data

    @property
    def vector_change(self) -> VectorChange:
        """Vector value change descriptor."""
        assert self.kind is TokenKind.CHANGE_VECTOR
        assert isinstance(self.data, VectorChange)
        return self.data

    @property
    def real_change(self) -> RealChange:
        """Real (float) value change descriptor."""
        assert self.kind is TokenKind.CHANGE_REAL
        assert isinstance(self.data, RealChange)
        return self.data

    @property
    def string_change(self) -> StringChange:
        "String value change descriptor."
        assert self.kind is TokenKind.CHANGE_STRING
        assert isinstance(self.data, StringChange)
        return self.data


class VCDParseError(Exception):
    """Catch-all error for any VCD parsing errors."""

    def __init__(self, loc: Location, msg: str) -> None:
        super().__init__(f'{loc.line}:{loc.column}: {msg}')
        self.loc = loc
        "Location within VCD file where error was detected."


HasReadinto = Union[io.BufferedIOBase, io.RawIOBase]


def tokenize(stream: HasReadinto, buf_size: Optional[int] = None) -> Iterator[Token]:
    """Parse VCD stream into tokens.

    The input stream must be opened in binary mode. E.g. with ``open(path, 'rb')``.

    """
    if buf_size is None:
        buf_size = io.DEFAULT_BUFFER_SIZE

    s = _TokenizerState(stream, bytearray(buf_size))

    try:
        while True:
            s.advance()
            yield _parse_token(s)
    except StopIteration:
        return


@dataclass
class _TokenizerState:
    stream: HasReadinto
    buf: bytearray
    pos: int = 0
    end: int = 0
    lineno: int = 1
    column: int = 0

    @property
    def loc(self) -> Location:
        return Location(self.lineno, self.column)

    def span(self, start: Location) -> Span:
        return Span(start, self.loc)

    def advance(self, raise_on_eof: bool = True) -> int:
        if self.pos < self.end:
            self.pos += 1
        else:
            n = self.stream.readinto(self.buf)
            if n:
                self.end = n - 1
                self.pos = 0
            elif raise_on_eof:
                raise StopIteration()
            else:
                return 0
        c = self.buf[self.pos]
        if c == 10:
            self.lineno += 1
            self.column = 1
        else:
            self.column += 1
        return self.buf[self.pos]

    def skip_ws(self) -> int:
        c = self.buf[self.pos]
        while c == 32 or 9 <= c <= 13:
            c = self.advance()
        return c

    def take_ws_after_kw(self, kw: str) -> None:
        if _is_ws(self.buf[self.pos]):
            self.advance()
        else:
            raise VCDParseError(self.loc, f'Expected whitespace after identifier ${kw}')

    def take_decimal(self) -> int:
        digits = []
        c = self.buf[self.pos]
        while 48 <= c <= 57:  # '0' <= c <= '9'
            digits.append(c)
            c = self.advance(raise_on_eof=False)
        if digits:
            return int(bytes(digits))
        else:
            raise VCDParseError(self.loc, 'Expected decimal value')

    def take_id_code(self) -> str:
        printables = []
        c = self.buf[self.pos]
        while 33 <= c <= 126:  # printable character
            printables.append(c)
            c = self.advance(raise_on_eof=False)
        if printables:
            return bytes(printables).decode('ascii')
        else:
            raise VCDParseError(self.loc, 'Expected id code')

    def take_identifier(self) -> str:
        # TODO: handle escaped identifiers
        identifier = []
        c = self.buf[self.pos]

        # Identifiers must start with letter or underscore
        if (
            65 <= c <= 90  # 'A' <= c <= 'Z'
            or 97 <= c <= 122  # 'a' - 'z'
            or c == 95  # '_'
        ):
            identifier.append(c)
            c = self.advance()
        else:
            raise VCDParseError(self.loc, 'Identifier must start with a-zA-Z_')

        while (
            48 <= c <= 57  # '0' - '9'
            or 65 <= c <= 90  # 'A' - 'Z'
            or 97 <= c <= 122  # 'a' - 'z'
            or c == 95  # '_'
            or c == 36  # '$'
            or c == 46  # '.' not in spec, but seen in the wild
        ):
            identifier.append(c)
            c = self.advance(raise_on_eof=False)

        return bytes(identifier).decode('ascii')

    def take_bit_index(self) -> Union[int, Tuple[int, int]]:
        self.skip_ws()
        index0 = self.take_decimal()
        index1: Optional[int]

        c = self.skip_ws()
        if c == 58:  # ':'
            self.advance()
            self.skip_ws()
            index1 = self.take_decimal()
        else:
            index1 = None

        c = self.skip_ws()
        if c == 93:  # ']'
            self.advance(raise_on_eof=False)
            if index1 is None:
                return index0
            else:
                return (index0, index1)
        else:
            raise VCDParseError(self.loc, 'Expected bit index to terminate with "]"')

    def take_to_end(self) -> str:
        chars = [
            self.buf[self.pos],  # $
            self.advance(),  # --> e
            self.advance(),  # --> n
            self.advance(),  # --> d
        ]
        while not (  # Check for 'd' 'n' 'e' '$'
            chars[-1] == 100
            and chars[-2] == 110
            and chars[-3] == 101
            and chars[-4] == 36
        ):
            chars.append(self.advance())

        if len(chars) > 4 and not _is_ws(chars[-5]):
            loc = Location(self.lineno, self.column - min(len(chars), 5))
            raise VCDParseError(loc, 'Expected whitespace before $end')

        return bytes(chars[:-5]).decode('ascii')

    def take_end(self) -> None:
        if (
            self.skip_ws() != 36  # '$'
            or self.advance() != 101  # 'e'
            or self.advance() != 110  # 'n'
            or self.advance() != 100  # 'd'
        ):
            raise VCDParseError(self.loc, 'Expected $end')


def _is_ws(c: int) -> bool:
    return c == 32 or 9 <= c <= 13


def _parse_token(s: _TokenizerState) -> Token:
    c = s.skip_ws()
    start = s.loc
    if c == 35:  # '#'
        # Parse time change
        s.advance()
        time = s.take_decimal()
        return Token(TokenKind.CHANGE_TIME, s.span(start), time)
    elif c == 48 or c == 49 or c == 122 or c == 90 or c == 120 or c == 88:
        # c in '01zZxX'
        # Parse scalar change
        scalar_value = chr(c)
        s.advance()
        id_code = s.take_id_code()
        return Token(
            TokenKind.CHANGE_SCALAR, s.span(start), ScalarChange(id_code, scalar_value)
        )
    elif c == 66 or c == 98:  # 'B' or 'b'
        # Parse vector change
        vector = []
        c = s.advance()
        while c == 48 or c == 49:  # '0' or '1'
            vector.append(c)
            c = s.advance()
        vector_value: Union[int, str]

        if c == 122 or c == 90 or c == 120 or c == 88:  # c in 'zZxX'
            vector.append(c)
            c = s.advance()
            while (
                c == 48 or c == 49 or c == 122 or c == 90 or c == 120 or c == 88
            ):  # c in '01zZxX'
                vector.append(c)
                c = s.advance()
            vector_value = bytes(vector).decode('ascii')
        else:
            vector_value = int(bytes(vector), 2)

        if not _is_ws(c):
            raise VCDParseError(s.loc, 'Expected whitespace after vector value')

        s.skip_ws()

        id_code = s.take_id_code()

        return Token(
            TokenKind.CHANGE_VECTOR, s.span(start), VectorChange(id_code, vector_value)
        )
    elif c == 82 or c == 114:  # 'R' or 'r'
        # Parse real change
        real_digits = []
        c = s.advance()

        while not _is_ws(c):
            real_digits.append(c)
            c = s.advance()

        try:
            real = float(bytes(real_digits))
        except ValueError:
            real_str = bytes(real_digits).decode("ascii")
            raise VCDParseError(start, f'Expected real value, got: {real_str}')

        s.skip_ws()

        id_code = s.take_id_code()

        return Token(TokenKind.CHANGE_REAL, s.span(start), RealChange(id_code, real))
    elif c == 83 or c == 115:  # 'S' or 's'
        chars = []
        c = s.advance()
        while not _is_ws(c):
            chars.append(c)
            c = s.advance()
        s.skip_ws()
        id_code = s.take_id_code()
        string_value = bytes(chars).decode('ascii')
        return Token(
            TokenKind.CHANGE_STRING, s.span(start), StringChange(id_code, string_value)
        )
    elif c == 36:  # '$'
        s.advance()
        kw = s.take_identifier()

        if kw == 'comment':
            s.take_ws_after_kw(kw)
            comment = s.take_to_end()
            return Token(TokenKind.COMMENT, s.span(start), comment)
        elif kw == 'date':
            s.take_ws_after_kw(kw)
            date_str = s.take_to_end()
            return Token(TokenKind.DATE, s.span(start), date_str)
        elif kw == 'enddefinitions':
            s.take_ws_after_kw(kw)
            s.take_end()
            return Token(TokenKind.ENDDEFINITIONS, s.span(start), None)
        elif kw == 'scope':
            s.take_ws_after_kw(kw)
            s.skip_ws()
            identifier = s.take_identifier()
            try:
                scope_type = ScopeType(identifier)
            except ValueError:
                raise VCDParseError(s.loc, f'Invalid $scope type: {identifier}')

            s.skip_ws()

            scope_ident = s.take_identifier()

            s.take_end()

            scope_decl = ScopeDecl(scope_type, scope_ident)

            return Token(TokenKind.SCOPE, s.span(start), scope_decl)
        elif kw == 'timescale':
            s.take_ws_after_kw(kw)
            s.skip_ws()
            mag_int = s.take_decimal()

            try:
                magnitude = TimescaleMagnitude(mag_int)
            except ValueError:
                valid_magnitudes = ', '.join(str(m.value) for m in TimescaleMagnitude)
                raise VCDParseError(
                    s.loc,
                    f'Invalid $timescale magnitude: {mag_int}. '
                    f'Must be one of: {valid_magnitudes}.',
                )

            s.skip_ws()
            unit_str = s.take_identifier()
            try:
                unit = TimescaleUnit(unit_str)
            except ValueError:
                valid_units = ', '.join(u.value for u in TimescaleUnit)
                raise VCDParseError(
                    s.loc,
                    f'Invalid $timescale unit: {unit_str}. '
                    f'Must be one of: {valid_units}.',
                )

            s.take_end()

            timescale = Timescale(magnitude, unit)
            return Token(TokenKind.TIMESCALE, s.span(start), timescale)
        elif kw == 'upscope':
            s.take_ws_after_kw(kw)
            s.take_end()
            return Token(TokenKind.UPSCOPE, s.span(start), None)
        elif kw == 'var':
            s.take_ws_after_kw(kw)
            s.skip_ws()
            type_str = s.take_identifier()
            try:
                type_ = VarType(type_str)
            except ValueError:
                valid_types = ', '.join(t.value for t in VarType)
                raise VCDParseError(
                    s.loc,
                    f'Invalid $var type: {type_str}. Must be one of: {valid_types}',
                )

            s.skip_ws()
            size = s.take_decimal()
            s.skip_ws()
            id_code = s.take_id_code()
            s.skip_ws()
            ident = s.take_identifier()

            bit_index: Union[None, int, Tuple[int, int]]
            c = s.skip_ws()
            if c == 91:  # '['
                s.advance()
                bit_index = s.take_bit_index()
            else:
                bit_index = None

            s.take_end()
            var_decl = VarDecl(type_, size, id_code, ident, bit_index)
            return Token(TokenKind.VAR, s.span(start), var_decl)
        elif kw == 'version':
            s.take_ws_after_kw(kw)
            version = s.take_to_end()
            return Token(TokenKind.VERSION, s.span(start), version)
        elif kw == 'dumpall':
            return Token(TokenKind.DUMPALL, s.span(start), None)
        elif kw == 'dumpoff':
            return Token(TokenKind.DUMPOFF, s.span(start), None)
        elif kw == 'dumpon':
            return Token(TokenKind.DUMPON, s.span(start), None)
        elif kw == 'dumpvars':
            return Token(TokenKind.DUMPVARS, s.span(start), None)
        elif kw == 'end':
            return Token(TokenKind.END, s.span(start), None)
        else:
            raise VCDParseError(s.loc, f'invalid keyword ${kw}')
    else:
        raise VCDParseError(s.loc, f'confused: {chr(c)}')



from enum import Enum
from typing import NamedTuple


class ScopeType(Enum):
    """Valid VCD scope types."""

    begin = 'begin'
    fork = 'fork'
    function = 'function'
    module = 'module'
    task = 'task'


class VarType(Enum):
    """Valid VCD variable types."""

    event = 'event'
    integer = 'integer'
    parameter = 'parameter'
    real = 'real'
    realtime = 'realtime'
    reg = 'reg'
    supply0 = 'supply0'
    supply1 = 'supply1'
    time = 'time'
    tri = 'tri'
    triand = 'triand'
    trior = 'trior'
    trireg = 'trireg'
    tri0 = 'tri0'
    tri1 = 'tri1'
    wand = 'wand'
    wire = 'wire'
    wor = 'wor'
    string = 'string'

    def __str__(self) -> str:
        return self.value


class TimescaleMagnitude(Enum):
    """Valid timescale magnitudes."""

    one = 1
    ten = 10
    hundred = 100


class TimescaleUnit(Enum):
    """Valid timescale units."""

    second = 's'
    millisecond = 'ms'
    microsecond = 'us'
    nanosecond = 'ns'
    picosecond = 'ps'
    femtosecond = 'fs'


class Timescale(NamedTuple):
    """Timescale magnitude and unit."""

    magnitude: TimescaleMagnitude
    unit: TimescaleUnit

    @classmethod
    def from_str(cls, s: str) -> 'Timescale':
        for unit in TimescaleUnit:
            if s == unit.value:
                mag = TimescaleMagnitude(1)
                break
        else:
            for mag in reversed(TimescaleMagnitude):
                mag_str = str(mag.value)
                if s.startswith(mag_str):
                    unit_str = s[len(mag_str) :].lstrip(' ')
                    unit = TimescaleUnit(unit_str)
                    break
            else:
                raise ValueError(f'Invalid timescale magnitude {s!r}')
        return Timescale(mag, unit)

    def __str__(self) -> str:
        return f'{self.magnitude.value} {self.unit.value}'



"""GTKWave save file generator.

This module provides tools for generating GTKWave save files.

`GTKWave`__ is an application for viewing VCD data. When opening a VCD file with
GTKWave, by default, no VCD variables (signals) are displayed. It is thus useful to have
an accompanying "save" file which configures various aspects of how GTKWave shows the
VCD data, including which variables are displayed, variable aliases, color information,
and more.

__ http://gtkwave.sourceforge.net

"""
import datetime
import math
import os
import time
import warnings
from contextlib import contextmanager
from enum import Enum, Flag, auto
from typing import IO, Any, Dict, Generator, List, Optional, Sequence, Tuple, Union


class GTKWFlag(Flag):
    """These are the valid GTKWave trace flags."""

    highlight = auto()
    "Highlight the trace item"

    hex = auto()  #: Hexadecimal data value representation
    dec = auto()  #: Decimal data value representation
    bin = auto()  #: Binary data value representation
    oct = auto()  #: Octal data value representation

    rjustify = auto()
    "Right-justify signal name/alias"
    invert = auto()
    reverse = auto()
    exclude = auto()
    blank = auto()  #: Used for blank, label, and/or analog height

    signed = auto()
    "Signed (2's compliment) data representation"
    ascii = auto()
    "ASCII character representation"

    collapsed = auto()  #: Used for closed groups
    ftranslated = auto()  #: Trace translated with filter file
    ptranslated = auto()  #: Trace translated with filter process

    analog_step = auto()  #: Show trace as discrete analog steps
    analog_interpolated = auto()  #: Show trace as analog with interpolation
    analog_blank_stretch = auto()  #: Used to extend height of analog data
    real = auto()  #: Real (floating point) data value representation
    analog_fullscale = auto()  #: Analog data scaled using full simulation time
    zerofill = auto()
    onefill = auto()
    closed = auto()

    grp_begin = auto()
    "Begin a group of signals"
    grp_end = auto()
    "End a group of signals"

    bingray = auto()
    graybin = auto()
    real2bits = auto()
    ttranslated = auto()

    popcnt = auto()
    "Show the population count, i.e. the number of set bits"

    fpdecshift = auto()


class GTKWColor(Enum):
    """The colors used by GTKWave.

    The `cycle` color is special and indicates the GTKWave should cycle through this
    list of colors, starting from the last selected color.

    """

    cycle = -1
    "Cycle between colors"
    normal = 0
    "Default color"
    red = 1
    orange = 2
    yellow = 3
    green = 4
    blue = 5
    indigo = 6
    violet = 7

    def _cycle(self):
        return GTKWColor((self.value + 1) % 8)


class GTKWSave:
    """Write GTKWave save files.

    This class provides methods for writing the various pieces of a GTKWave save file. A
    GTKWave save file compliments a VCD dump file with dump file specific configuration
    GTKWave uses to display the dump file.

    A GTKWave save file is line-oriented ASCII text. Each line consists of a single
    configuration directive. All directives are optional.

    Some directives, such as :meth:`dumpfile()`, are for general GTKWave configuration.
    These general directives may be added anywhere in the save file and in any order
    relative to other directives. Directives may also be duplicated--the last one added
    will be used by GTKWave.

    The :meth:`trace()`, :meth:`trace_bits()`, :meth:`group()`, and :meth:`blank`
    directives add signals to the "Signals" list which are traced in the "Waves" frame.
    The order in which these signal traces are added determines the order in GTKWave.

    """

    def __init__(self, savefile: IO[str]) -> None:
        self.file = savefile
        self.path = getattr(savefile, 'name', None)
        self._flags = GTKWFlag(0)
        self._color_stack = [GTKWColor.normal]
        self._filter_files: List[str] = []
        self._filter_procs: List[str] = []

    def _p(self, *args: object, **kwargs) -> None:
        print(*args, file=self.file, **kwargs)

    def _set_flags(self, flags: GTKWFlag) -> None:
        if flags != self._flags:
            self._p(f'@{flags.value:x}')
            self._flags = flags

    def _set_color(self, color: Optional[Union[GTKWColor, str, int]]) -> None:
        if color is not None:
            if not isinstance(color, GTKWColor):
                warnings.warn(
                    'Using str and int for colors is deprecated. '
                    'Use vcd.gtkw.GTKWColor instead.',
                    DeprecationWarning,
                    stacklevel=2,
                )
                if isinstance(color, str):
                    color = GTKWColor.__members__[color]
                else:
                    assert isinstance(color, int)
                    color = GTKWColor(color)
                assert isinstance(color, GTKWColor)

            if color == GTKWColor.cycle:
                self._color_stack[-1] = self._color_stack[-1]._cycle()
            else:
                self._color_stack[-1] = color
            self._p(f'[color] {self._color_stack[-1].value}')

    def _set_translate_filter_file(self, filter_path: Optional[str]) -> None:
        if filter_path:
            try:
                filter_id = 1 + self._filter_files.index(filter_path)
            except ValueError:
                self._filter_files.append(filter_path)
                filter_id = len(self._filter_files)
            self._p(f'^{filter_id} {filter_path}')

    def _set_translate_filter_proc(self, proc_path: Optional[str]) -> None:
        if proc_path:
            try:
                filter_id = 1 + self._filter_procs.index(proc_path)
            except ValueError:
                self._filter_procs.append(proc_path)
                filter_id = len(self._filter_procs)
            self._p(f'^>{filter_id} {proc_path}')

    def comment(self, *comments: Sequence[str]) -> None:
        """Add comment line(s) to save file."""
        for comment in comments:
            self._p('[*]', comment)

    def dumpfile(self, dump_path: str, abspath: bool = True) -> None:
        """Add VCD dump file path to save file.

        The `[dumpfile]` must be in the save file in order to only have to specify the
        save file on the `gtkwave` command line. I.e.:

            $ gtkwave my.gtkw

        If the `[dumpfile]` is not present in the save file, both the dump and save
        files must be specified to `gtkwave`:

            $ gtkwave my.vcd my.gtkw

        :param dump_path: path to VCD dump file or None to produce special "(null)"
                          value in the save file.
        :param bool abspath: convert *dump_path* to an absolute path.

        """
        if dump_path is None:
            self._p('[dumpfile] (null)')
        else:
            if abspath:
                dump_path = os.path.abspath(dump_path)
            self._p(f'[dumpfile] "{dump_path}"')

    def dumpfile_mtime(
        self,
        mtime: Optional[Union[float, time.struct_time, datetime.datetime]] = None,
        dump_path: Optional[str] = None,
    ) -> None:
        """Add dump file modification time to save file.

        Configuring the dump file's modification time is optional.

        """
        time_format = '%a %b %d %H:%M:%S %Y'
        if mtime is None:
            assert isinstance(dump_path, str)
            mtime = os.stat(dump_path).st_mtime
        if isinstance(mtime, float):
            mtime = time.gmtime(mtime)
        if isinstance(mtime, time.struct_time):
            mtime_str = time.strftime(time_format, mtime)
        elif isinstance(mtime, datetime.datetime):
            mtime_str = mtime.strftime(time_format)
        else:
            raise TypeError(f'Invalid mtime type ({type(mtime)})')
        self._p(f'[dumpfile_mtime] "{mtime_str}"')

    def dumpfile_size(
        self, size: Optional[int] = None, dump_path: Optional[str] = None
    ) -> None:
        """Add dump file size annotation to save file.

        Configuring the dump file's size is optional.

        """
        if size is None:
            assert isinstance(dump_path, str)
            size = os.stat(dump_path).st_size
        self._p(f'[dumpfile_size] {size}')

    def savefile(self, save_path: Optional[str] = None, abspath: bool = True) -> None:
        """Add the path of the save file to the save file.

        With no parameters, the output file's name will be used.

        Configuring the `[savefile]` is optional.

        :param save_path: path to this save file. None will use the output file's path.
        :param bool abspath: determines whether to make the path absolute.

        """
        if save_path is None and self.path is None:
            self._p('[savefile] (null)')
        else:
            if save_path is None:
                save_path = self.path
            if abspath and save_path is not None:
                save_path = os.path.abspath(save_path)
            self._p(f'[savefile] "{save_path}"')

    def timestart(self, timestamp: int = 0) -> None:
        """Add simulation start time to the save file."""
        self._p(f'[timestart] {timestamp}')

    def zoom_markers(
        self, zoom: float = 0.0, marker: int = -1, **kwargs: Dict[str, int]
    ) -> None:
        """Set zoom, primary marker, and markers 'a' - 'z'."""
        self._p(
            f'*{zoom:.6f} {marker}',
            *[kwargs.get(k, -1) for k in 'abcdefghijklmnopqrstuvwxyz'],
        )

    def size(self, width: int, height: int) -> None:
        """Set GTKWave window size."""
        self._p(f'[size] {width} {height}')

    def pos(self, x: int = -1, y: int = -1) -> None:
        """Set GTKWave window position."""
        self._p(f'[pos] {x} {y}')

    def treeopen(self, tree: str) -> None:
        """Start with *tree* open in Signal Search Tree (SST).

        GTKWave specifies tree paths with a trailing '.'. The trailing '.' will
        automatically be added if it is omitted in the *tree* parameter.

        :param str tree: scope/path/tree to be opened in GTKWave's SST frame.

        """
        if tree[-1] == '.':
            self._p(f'[treeopen] {tree}')
        else:
            self._p(f'[treeopen] {tree}.')

    def signals_width(self, width: int) -> None:
        """Set width of Signals frame."""
        self._p(f'[signals_width] {width}')

    def sst_expanded(self, is_expanded: bool) -> None:
        """Set whether Signal Search Tree (SST) frame is expanded."""
        self._p(f'[sst_expanded] {int(bool(is_expanded))}')

    def pattern_trace(self, is_enabled: bool) -> None:
        """Enable/disable pattern trace."""
        self._p(f'[pattern_trace] {int(bool(is_enabled))}')

    @contextmanager
    def group(
        self, name: str, closed: bool = False, highlight: bool = False
    ) -> Generator[None, None, None]:
        """Contextmanager helper for :meth:`begin_group` and :meth:`end_group`.

        This context manager starts a new group of signal traces and ends the group when
        leaving the `with` block. E.g.:

            >>> import io
            >>> gtkw = GTKWSave(io.StringIO())
            >>> with gtkw.group('mygroup'):
            ...     gtkw.trace('a.b.x')
            ...     gtkw.trace('a.b.y')
            ...     gtkw.trace('a.b.z')

        :param str name: the name/label of the trace group.
        :param bool closed: group should be closed at GTKWave startup.
        :param bool highlight: group should be highlighted at GTKWave startup.

        """
        self.begin_group(name, closed, highlight)
        try:
            yield None
        finally:
            self.end_group(name, closed, highlight)

    def begin_group(
        self, name: str, closed: bool = False, highlight: bool = False
    ) -> None:
        """Begin a new signal trace group.

        Consider using :meth:`group()` instead of :meth:`begin_group()` and
        :meth:`end_group()`.

        :param str name: the name/label of the trace group.
        :param bool closed: group should be closed at GTKWave startup.
        :param bool highlight: group should be highlighted at GTKWave startup.

        """
        flags = GTKWFlag.grp_begin | GTKWFlag.blank
        if closed:
            flags |= GTKWFlag.closed
        if highlight:
            flags |= GTKWFlag.highlight
        self._set_flags(flags)
        self._p(f'-{name}')
        self._color_stack.append(GTKWColor.normal)

    def end_group(
        self, name: str, closed: bool = False, highlight: bool = False
    ) -> None:
        """End a signal trace group.

        This call must match with a prior call to :meth:`begin_group(). Consider using
        :meth:`group()` instead of :meth:`begin_group()` and :meth:`end_group()`.

        :param str name: the name/label of the trace group.
        :param bool closed: group should be closed at GTKWave startup.
        :param bool highlight: group should be highlighted at GTKWave startup.

        """
        flags = GTKWFlag.grp_end | GTKWFlag.blank
        if closed:
            flags |= GTKWFlag.closed | GTKWFlag.collapsed
        if highlight:
            flags |= GTKWFlag.highlight
        self._set_flags(flags)
        self._p(f'-{name}')
        self._color_stack.pop(-1)

    def blank(
        self, label: str = '', analog_extend: bool = False, highlight: bool = False
    ) -> None:
        """Add blank or label to trace signals list.

        :param str label: Optional label for the blank.
        :param bool analog_extend: extend the height of an immediately preceding analog
                                   trace signal.
        :param bool highlight: blank should be highlighted at GTKWave startup.

        """
        flags = GTKWFlag.blank
        if analog_extend:
            flags |= GTKWFlag.analog_blank_stretch
        if highlight:
            flags |= GTKWFlag.highlight
        self._set_flags(flags)
        self._p(f'-{label}')

    def trace(
        self,
        name: str,
        alias: Optional[str] = None,
        color: Optional[Union[GTKWColor, str, int]] = None,
        datafmt: str = 'hex',
        highlight: bool = False,
        rjustify: bool = True,
        extraflags: Union[GTKWFlag, Optional[Sequence[str]]] = GTKWFlag(0),
        translate_filter_file: Optional[str] = None,
        translate_filter_proc: Optional[str] = None,
    ) -> None:
        """Add signal trace to save file.

        :param str name: fully-qualified name of signal to trace.
        :param str alias: optional alias to display instead of the *name*.
        :param GTKWColor color: optional color to use for the signal's trace.
        :param str datafmt: the format used for data display. Must be one of 'hex',
                            'dec', 'bin', 'oct', 'ascii', 'real', or 'signed'.
        :param bool highlight: trace should be highlighted at GTKWave startup.
        :param bool rjustify: trace name/alias should be right-justified.
        :param GTKWFlag extraflags: extra flags to apply to the trace.
        :param str translate_filter_file: path to translate filter file.
        :param str translate_filter_proc: path to translate filter executable.

        .. Note::

            GTKWave versions <= 3.3.64 require vector signal names to have a bit range
            suffix. For example, an 8-bit vector variable "module.myint" would be known
            by GTKWave as "module.myint[7:0]".

            GTKWave versions after 3.3.64 do not use bit-range suffixes.

        """
        if datafmt not in ['hex', 'dec', 'bin', 'oct', 'ascii', 'real', 'signed']:
            raise ValueError(f'Invalid datafmt ({datafmt})')
        flags = GTKWFlag.__members__[datafmt]
        if isinstance(extraflags, GTKWFlag):
            flags |= extraflags
        else:
            warnings.warn(
                'Using Optional[Sequence[str]] for extraflags is deprecated. '
                'Use vcd.gtkw.GTKWFlag instead.',
                DeprecationWarning,
            )
            if extraflags is not None:
                for flag in GTKWFlag:
                    if flag.name in extraflags:
                        flags |= flag
        if highlight:
            flags |= GTKWFlag.highlight
        if rjustify:
            flags |= GTKWFlag.rjustify
        if translate_filter_file:
            flags |= GTKWFlag.ftranslated
        if translate_filter_proc:
            flags |= GTKWFlag.ptranslated
        self._set_flags(flags)
        self._set_color(color)
        self._set_translate_filter_file(translate_filter_file)
        self._set_translate_filter_proc(translate_filter_proc)
        if alias:
            self._p(f'+{{{alias}}} ', end='')
        self._p(name)

    @contextmanager
    def trace_bits(
        self,
        name: str,
        alias: Optional[str] = None,
        color: Optional[Union[str, int]] = None,
        datafmt: str = 'hex',
        highlight: bool = False,
        rjustify: bool = True,
        extraflags: Union[GTKWFlag, Optional[Sequence[str]]] = GTKWFlag(0),
        translate_filter_file: Optional[str] = None,
        translate_filter_proc: Optional[str] = None,
    ) -> Generator[None, None, None]:
        """Contextmanager for tracing bits of a vector signal.

        This allows each individual bit of a vector signal to have its own trace (and
        trace configuration).

            >>> import io
            >>> gtkw = GTKWSave(io.StringIO())
            >>> name = 'mod.myint'
            >>> with gtkw.trace_bits(name):
            ...     gtkw.trace_bit(0, name)
            ...     gtkw.trace_bit(1, name)
            ...     gtkw.trace_bit(2, name)
            ...     gtkw.trace_bit(3, name, 'special', color=GTKWColor.yellow)

        :param str name: fully-qualified name of the vector variable to trace.
        :param str alias: optional alias to display instead of *name*.
        :param int color: optional trace color.
        :param str datafmt: format for data display.
        :param bool highlight: trace should be highlighted at GTKWave startup.
        :param bool rjustify: trace name/alias should be right-justified.
        :param GTKWFlag extraflags: extra flags to apply to the trace.
        :param str translate_filter_file: path to translate filter file.
        :param str translate_filter_proc: path to translate filter executable.

        """
        self.trace(
            name,
            alias,
            color,
            datafmt,
            highlight,
            rjustify,
            extraflags,
            translate_filter_file,
            translate_filter_proc,
        )
        flags = GTKWFlag.bin
        if isinstance(extraflags, GTKWFlag):
            flags |= extraflags
        else:
            warnings.warn(
                'Using Optional[Sequence[str]] for extraflags is deprecated. '
                'Use vcd.gtkw.GTKWFlag instead.',
                DeprecationWarning,
            )
            if extraflags is not None:
                for flag in GTKWFlag:
                    if flag.name in extraflags:
                        flags |= flag
        if highlight:
            flags |= GTKWFlag.highlight
        if rjustify:
            flags |= GTKWFlag.rjustify
        self._set_flags(flags)
        try:
            yield None
        finally:
            flags = GTKWFlag.blank | GTKWFlag.grp_end | GTKWFlag.collapsed
            if highlight:
                flags |= GTKWFlag.highlight
            self._set_flags(flags)
            self._p('-group_end')

    def trace_bit(
        self,
        index: int,
        name: str,
        alias: Optional[str] = None,
        color: Optional[Union[GTKWColor, str, int]] = None,
    ) -> None:
        """Trace individual bit of vector signal.

        This is meant for use in conjunction with :meth:`trace_bits()`.

        :param int index: index of bit
        :param str name: name of parent vector signal.
        :param str alias: optional alias to display for bit.
        :param int color: optional color for bit's trace.

        """
        self._set_color(color)
        if alias:
            self._p(f'+{{{alias}}} ', end='')
        self._p(f'({index}){name}')


TranslationType = Union[
    Tuple[Union[int, str], str], Tuple[Union[int, str], str, Union[str, int]]
]


def make_translation_filter(
    translations: Sequence[Tuple[Any, ...]],
    datafmt: str = 'hex',
    size: Optional[int] = None,
) -> str:
    """Create translation filter.

    The returned translation filter string that can be written to a translation filter
    file usable by GTKWave.

    :param translations:

        Sequence of 2-tuples `(value, alias)` or 3-tuples `(value, alias, color)`.

    :param str datafmt:

        Format to apply to the translation values. This *datafmt* must match the
        *datafmt* used with :meth:`GTKWSave.trace()`, otherwise these translations will
        not be matched by GTKWave.

    :returns: Translation filter string suitable for writing to a translation filter
              file.

    """
    if datafmt == "hex":
        assert isinstance(size, int)
        value_format = f'0{int(math.ceil(size / 4))}x'
    elif datafmt == 'oct':
        assert isinstance(size, int)
        value_format = f'0{int(math.ceil(size / 3))}o'
    elif datafmt in ['dec', 'signed']:
        value_format = 'd'
    elif datafmt == 'bin':
        assert isinstance(size, int)
        value_format = f'0{size}b'
    elif datafmt == 'real':
        value_format = '.16g'
    elif datafmt == 'ascii':
        value_format = ""
        ascii_translations = []
        for translation in translations:
            value = translation[0]
            rest = list(translation[1:])
            if isinstance(value, int):
                value = bytes((value,)).decode('ascii')
            elif not isinstance(value, str):
                raise TypeError(f'Invalid type ({type(value)}) for ascii translation')
            elif len(value) != 1:
                raise ValueError(f'Invalid ascii string "{value}"')
            ascii_translations.append(tuple([value] + rest))
        translations = ascii_translations
    else:
        raise ValueError(f'invalid datafmt ({datafmt})')

    lines = []

    for translation in translations:
        if len(translation) == 2:
            value, label = translation
            color = None
        else:
            value, label, color = translation

        if datafmt in ['hex', 'oct', 'bin']:
            assert isinstance(size, int)
            max_val = 1 << size
            if -value > (max_val >> 1) or value >= max_val:
                raise ValueError(f'Value ({value}) not representable in {size} bits')
            if value < 0:
                # Two's compliment treatment
                value += 1 << size

        value_str = format(value, value_format)

        if color is None:
            lines.append(f'{value_str} {label}')
        else:
            lines.append(f'{value_str} ?{color}?{label}')

    return '\n'.join(lines)


def decode_flags(flags: Union[str, int]) -> List[str]:
    """Decode hexadecimal flags from GTKWave save file into flag names.

    This is useful for understanding what, for example "@802022" means when inspecting a
    GTKWave save file.

    :param flags: Hexadecimal flags from GTKWave save file; either as an integer or
                  string with hexadecimal characters.
    :returns: List of flag names

    """
    if isinstance(flags, str):
        decoded = GTKWFlag(int(flags.lstrip('@'), 16))
    else:
        decoded = GTKWFlag(flags)
    return [flag.name for flag in GTKWFlag if flag & decoded]


def spawn_gtkwave_interactive(
    dump_path: str, save_path: str, quiet: bool = False
) -> None:  # pragma: no cover
    """Spawn gtkwave process in interactive mode.

    A process pipeline is constructed such that the contents of the VCD dump file at
    *dump_path* are displayed interactively as the dump file is being written (i.e. with
    :class:`~vcd.writer.VCDWriter`.

    The process pipeline built is approximately equivalent to::

        $ tail -f dump_path | shmidcat | gtkwave -vI save_path

    The ``tail``, ``shmidcat``, and ``gtkwave`` executables must be found in ``$PATH``.

    .. Warning::

        This function does not work on Windows.

    .. Note::

        A child python process of the caller will remain running until the GTKWave
        window is closed. This process ensures that the various other child processes
        are properly reaped.

    :param str dump_path: path to VCD dump file. The dump file must exist, but be empty.
    :param str save_path: path to GTKWave save file. The save file will be read
                          immediately by GTKWave and thus must be completely written.
    :param bool quiet: quiet GTKWave's output by closing its `stdout` and `stderr` file
                       descriptors.

    """
    import signal

    stdin_fd, stdout_fd, stderr_fd = 0, 1, 2

    if not os.fork():
        shmidcat_rd_fd, tail_wr_fd = os.pipe()

        tail_pid = os.fork()
        if not tail_pid:
            os.close(shmidcat_rd_fd)
            os.dup2(tail_wr_fd, stdout_fd)
            os.execlp('tail', 'tail', '-n', '+0', '-f', dump_path)

        os.close(tail_wr_fd)
        gtkwave_rd_fd, shmidcat_wr_fd = os.pipe()

        shmidcat_pid = os.fork()
        if not shmidcat_pid:
            os.close(gtkwave_rd_fd)
            os.dup2(shmidcat_rd_fd, stdin_fd)
            os.dup2(shmidcat_wr_fd, stdout_fd)
            os.execlp('shmidcat', 'shmidcat')

        os.close(shmidcat_rd_fd)
        os.close(shmidcat_wr_fd)

        gtkwave_pid = os.fork()
        if not gtkwave_pid:
            os.dup2(gtkwave_rd_fd, stdin_fd)
            if quiet:
                devnull = open(os.devnull, 'w')
                os.dup2(devnull.fileno(), stdout_fd)
                os.dup2(devnull.fileno(), stderr_fd)
            os.execlp('gtkwave', 'gtkwave', '--vcd', '--interactive', save_path)

        # The first forked process exists to do this cleanup...
        os.waitpid(gtkwave_pid, 0)
        os.kill(tail_pid, signal.SIGTERM)
        os.kill(shmidcat_pid, signal.SIGTERM)
        os._exit(0)



"""Value Change Dump (VCD) file support.

.. autosummary::

    ~vcd.writer.VCDPhaseError
    ~vcd.writer.VCDWriter
    ~vcd.reader.tokenize

"""
from .reader import tokenize
from .writer import VCDPhaseError, VCDWriter

__all__ = ('VCDWriter', 'VCDPhaseError', 'tokenize')
